"""Yaml Parser Module."""
import yaml
import sys
from buzio import console


class Config():
    """Configuration Class.

    Get and parse config from yaml configuration file
    defined in CABRITA_PATH environment settings or
    passed as parameter during initialization.
    """

    def __init__(self, path):
        """Init Class."""
        self.path = path

    def get_config(self):
        """Convert configuration yaml file to python dict."""
        try:
            with open(self.path, 'r') as file:
                data = yaml.load(file.read())
            return data
        except IOError as exc:
            console.error("Cannot open file: {}".format(exc))
            sys.exit(1)
        except yaml.YAMLError as exc:
            console.error("Cannot read file: {}".format(exc))
            sys.exit(1)
        except Exception as exc:
            console.error("Error: {}".format(exc))
            sys.exit(1)



"""Cabrita main module."""
import click
from cabrita import __version__
from buzio import console
from cabrita.parser import Config
from cabrita.dash import Dashboard
from cabrita.versions import check_version


@click.command()
@click.option(
    '--path',
    envvar="CABRITA_PATH",
    prompt='Inform full path to config file',
    help='Full path for configuration file.',
    type=click.Path())
def run(path):
    """Run main command for cabrita.

    1. Check version
    2. Import configuration file
    3. Run dashboard.
    """
    console.box("Cabrita v{}".format(__version__))
    check_version()
    console.info("Loading Configuration...")
    config = Config(path).get_config()
    dash = Dashboard(config)
    console.info("Reading data...")
    dash.run()


if __name__ == "__main__":
    run()



"""Utils module."""
import os
import sys
import yaml
import re
import subprocess
from buzio import console


def get_yaml(path, file):
    """Open yaml file."""
    try:
        dcfile = os.path.join(path, file)
        with open(dcfile, 'r') as file:
            data = yaml.load(file.read())
        return data
    except IOError as exc:
        console.error("Cannot open file: {}".format(exc))
        sys.exit(1)
    except yaml.YAMLError as exc:
        console.error("Cannot read file: {}".format(exc))
        sys.exit(1)
    except Exception as exc:
        console.error("Error: {}".format(exc))
        sys.exit(1)


def run_command(
        task,
        get_stdout=False,
        run_stdout=False):
    """Run subprocess command.

    Args:
        task (string): the command to run
        get_stdout (bool, optional): capture stdout
        run_stdout (bool, optional): capture and run stdout

    Returns:
        Bool or String: Bool if task was executed or stdout
    """
    try:
        if run_stdout:
            command = subprocess.check_output(task, shell=True)

            if not command:
                return False

            ret = subprocess.call(command, shell=True)

        elif get_stdout is True:
            ret = subprocess.check_output(task, shell=True)
        else:
            ret = subprocess.call(
                task,
                shell=True,
                stderr=subprocess.STDOUT
            )

        if ret != 0 and not get_stdout:
            return False
    except KeyboardInterrupt:
        raise KeyboardInterrupt
    except BaseException:
        return False

    return True if not get_stdout else ret.decode('utf-8')


def get_path(path, base_path):
    """Return real path from string.

    Converts environment variables to path
    Converts relative path to full path
    """
    if "$" in path:
        s = re.search("\${(\w+)}", path)
        if s:
            env = s.group(1)
            name = os.environ.get(env)
            path_list = [
                part if "$" not in part else name
                for part in path.split("/")
            ]
            path = os.path.join(*path_list)
        else:
            console.error(
                "Cant find path for {}".format(path)
            )
            return False
    if "." in path:
        list_path = os.path.join(base_path, path)
        path = os.path.abspath(list_path)
    return path



"""Cabrita Dashboard module."""
import os
import psutil
import re
import sys
import requests
import json
import datetime
import threading
from raven import Client
from blessed import Terminal
from buzio import formatStr, console
from cabrita import gui
from tabulate import tabulate
from time import sleep
from cabrita.utils import run_command, get_yaml, get_path

UP = u'▲'
DOWN = u'▼'

ARROW_UP = u"↑"
ARROW_DOWN = u"↓"

client = None
if os.getenv('CABRITA_SENTRY_DSN'):
    client = Client(os.getenv('CABRITA_SENTRY_DSN'))


class Dashboard():
    """Dashboard class."""

    def __init__(self, config):
        """Class initialization.

        Args:
            config (dict): config data from yaml file.
        """
        self.data = None
        self.services = None
        self.config = config
        self.log = None
        self.included_services = []
        self.last_git_check = None
        self.can_check_git = False
        self.cache_git = {}
        self.repo = None

    def get_compose_data(self):
        """Get data from docker-compose.yml.

        Convert services key in docker-compose.yml
        to python dict stored in self.services
        """
        self.path = self.config['docker-compose']['path']
        if "$" in self.path:
            converted_path_list = []
            path_list = os.path.split(self.path)
            for p in path_list:
                env = p
                if "$" in p:
                    s = re.search(r"\$(\w+)", p)
                    if s:
                        env = os.environ.get(s.group(1), None)
                        if not env:
                            console.error(
                                "Can't resolve {}".format(
                                    s.group(1)))
                            sys.exit(1)
                converted_path_list.append(env)
            self.path = os.path.join(*converted_path_list)
        self.data = get_yaml(
            path=self.path,
            file=self.config['docker-compose']['name']
        )
        self.services = sorted(self.data['services'])

    def get_data_from_version(self):
        """Set instance properties.

        Instance properties are set based on
        configuration version indicated in yaml file:

        Version 0: testing version (deprecated)
        Version 1: current version
        """
        version = self.config.get('version', 0)
        if version == 0:
            self.categories = self.config.get('categories', [])
            self.small_list = self.config.get('infra', [])
            self.interval = 2
            self.big_name = "Docker Services"
            self.small_name = "Services"
            self.layout = "horizontal"
            self.check_ngrok = True
            self.ignore = self.config.get('ignore', [])
            self.fetch = 30
        elif version == 1:
            self.boxes = self.config['box']
            self.interval = self.config.get('interval', 2)
            self.layout = self.config.get('layout', 'horizontal')
            self.check_ngrok = self.config.get('check', {}).get('ngrok', False)
            self.ignore = self.config.get('ignore', [])
            self.fetch = self.config.get('fetch_each', 30)

    def _make_box(self, box):
        """Generate box data.

        Iterate all services from docker-compose.yml, ignoring:
            1. Already processed services (in previous processed boxes)
            2. Ignored services (from 'ignore' list parameter)
            2. Recognized as category (from 'categories' list parameter)

        Check if service must be inside this box, using:
            1. 'filter' list parameter
            2. 'list_only' list parameter
            3. 'catch_all' parameter (if this is the last box)

        If service must be inside this box:
            1. get formatted service name (by status)
            2. get formatted service status (by status)
            3. Show git branch info (from 'show_git' parameter)
            4. Iterate all possible category services
               derived from this service

        Args:
            box (dict): box to process

        Returns
        -------
            obj: gui.Text instance

        """
        table_lines = []
        table_header = ['Service']
        show_git = box.get('show_git', True)
        categories = box.get('categories', [])
        list_only = box.get('list_only', [])
        box_name = box.get('name', box)
        box_filter = box.get('filter', "")
        for key in self.services:
            if key in self.included_services:
                continue
            jump = False
            for i in self.ignore:
                if i.lower() in key.lower():
                    jump = True
            if box_filter and box_filter.lower() not in key.lower():
                jump = True
            if jump:
                continue
            for name in categories:
                if name.lower() in key.lower():
                    jump = True
            if jump:
                continue
            if list_only:
                jump = True
                for s in list_only:
                    if s.lower() in key.lower():
                        jump = False
                if jump:
                    continue
            table_header = ["Service", "Status"]
            if show_git:
                table_header += ["Git"]
            table_header += categories
            table_data = [
                self.get_service_name(key),
                self._check_server(name=key)
            ]
            if show_git:
                table_data.append(self._get_git_status(key, box))
            self.included_services.append(key)
            for cat in categories:
                found = False
                for search in self.services:
                    if "_" in key:
                        k = key.lower().split("_")[0]
                    elif "-" in key:
                        k = key.lower().split("-")[0]
                    else:
                        k = key.lower()
                    if k in search.lower() and cat.lower() in search.lower():
                        table_data.append(self._check_server(search))
                        self.included_services.append(search)
                        found = True
                        continue
                if not found:
                    table_data.append(
                        formatStr.info("--", theme="dark", use_prefix=False)
                    )

            table_lines.append(table_data)

        text = tabulate(
            table_lines,
            table_header
        )
        return gui.Text(text, color=6, border_color=5, title=box_name)

    def get_service_name(self, key):
        """Return service name.

        Return formatted service name,
        based on docker inspect information.

        Args:
            key (string): Service name

        Returns
        -------
            string: formatted string for key

        """
        try:
            ret = run_command(
                'docker inspect {} 2>/dev/null'.format(key),
                get_stdout=True
            )
            if ret:
                ret = json.loads(ret)[0]
                if ret['State']['Running']:
                    return formatStr.info(key, use_prefix=False)
            return formatStr.info(key, theme="dark", use_prefix=False)
        except KeyboardInterrupt:
            raise KeyboardInterrupt
        except BaseException as exc:
            return formatStr.info(key, theme="error", use_prefix=False)

    def get_layout(self, term):
        """Generate the dashboard layout.

        First, get the logs, machine info and docker-compose/ngrok status
        Then generate layout for each box configured in yaml file.

        Args:
            term (obj): Python Terminal instance

        Returns
        -------
            obj: HSplit or VSplit instance

        """
        log = self.get_log()
        info = self.get_info()
        check_status = self.get_check_status()
        large_boxes = []
        small_boxes = []
        last_box = None
        for box in self.boxes:
            current_box = self.boxes[box]
            if current_box.get('catch_all', False):
                last_box = current_box
                continue
            size = current_box.get('size', 'small')
            if size == "small":
                small_boxes.append(self._make_box(current_box))
            else:
                large_boxes.append(self._make_box(current_box))
        size = last_box.get('size', 'small')
        if size == "small":
            small_boxes.insert(0, self._make_box(last_box))
        else:
            large_boxes.insert(0, self._make_box(last_box))

        sm = None
        lg = None
        if large_boxes:
            lg = large_boxes
        st = gui.VSplit(
            log,
            gui.VSplit(
                check_status,
                info
            )
        )
        if small_boxes:
            sm = gui.HSplit(*small_boxes, st)
        else:
            sm = st
        if self.layout == "horizontal":
            func = gui.HSplit
        else:
            func = gui.VSplit
        if not lg:
            ui = func(sm, terminal=term, main=True)
        else:
            ui = func(*lg, sm, terminal=term, main=True)

        return ui

    def get_info(self):
        """Get machine info using PSUtil.

        Returns
        -------
            obj: gui.HSplit instance

        """
        cpu_percent = round(psutil.cpu_percent(interval=None) * 10, 0) / 10
        free_memory = int(psutil.virtual_memory().available / 1024 / 1024)
        total_memory = int(psutil.virtual_memory().total / 1024 / 1024)
        memory_percent = (free_memory / total_memory) * 100
        free_space = round(psutil.disk_usage("/").free / 1024 / 1024 / 1024, 1)
        total_space = round(psutil.disk_usage(
            "/").total / 1024 / 1024 / 1024, 1)
        space_percent = (free_space / total_space) * 100

        if memory_percent > 100:
            memory_percent = 100

        if space_percent > 100:
            space_percent = 100

        if cpu_percent <= 50:
            cpu_color = 2
        elif cpu_percent <= 70:
            cpu_color = 3
        else:
            cpu_color = 1

        if memory_percent <= 20:
            memory_color = 1
        elif memory_percent <= 50:
            memory_color = 3
        else:
            memory_color = 2

        if space_percent <= 20:
            space_color = 1
        elif space_percent <= 50:
            space_color = 3
        else:
            space_color = 2

        widget = gui.HSplit(
            gui.HGauge(
                val=cpu_percent,
                color=cpu_color,
                border_color=cpu_color,
                title="CPU:{}%".format(cpu_percent)
            ),
            gui.HGauge(
                val=memory_percent,
                color=memory_color,
                border_color=memory_color,
                title="Free Mem:{}M".format(free_memory)
            ),
            gui.HGauge(
                val=space_percent,
                color=space_color,
                border_color=space_color,
                title="Free Space:{}Gb".format(free_space)
            )
        )
        return widget

    def get_log(self):
        """Generate logs.

        Returns
        -------
            obj: gui.Log instance

        """
        if not self.log:
            # First time checks
            self.log = gui.Log(
                date_format=self.config['logging']['date_format'],
                title="Log",
                color=6,
                border_color=5
            )
            self.log.info("Cabrita has started.")
            self.log.info("Press CTRL-C to end.")

        # Unwatched services
        if self.included_services:
            name_list = [
                service
                for service in self.services
                if service not in self.included_services and service not in self.ignore
            ]
            for name in name_list:
                self.log.warn("Not watched: {}".format(name))
        return self.log

    def get_check_status(self):
        """Check Docker-Compose and Ngrok status.

        Returns
        -------
            obj: gui.Text

        """
        # Check docker-compose.yml status
        ret = run_command(
            "cd {} && git fetch && git status -bs --porcelain".format(
                self.path),
            get_stdout=True
        )
        if not ret:
            text = formatStr.warning(
                "Can't find Docker-Compose status.\n",
                use_prefix=False)
        elif 'behind' in ret:
            text = formatStr.error(
                'Docker-Compose is OUTDATED.\n',
                use_prefix=False)
        else:
            text = formatStr.success(
                'Docker-Compose is up-to-date.\n',
                use_prefix=False)
        # Check Ngrok
        if self.check_ngrok:
            try:
                ret = requests.get(
                    "http://127.0.0.1:4040/api/tunnels", timeout=1)
                if ret.status_code == 200:
                    text += formatStr.success("Ngrok status: running",
                                              use_prefix=False)
                else:
                    text += formatStr.error("Ngrok status: ERROR",
                                            use_prefix=False)
            except KeyboardInterrupt:
                raise KeyboardInterrupt
            except BaseException:
                text += formatStr.error("Ngrok status: NOT RUNNING",
                                        use_prefix=False)
        return gui.Text(text, border_color=5, title="Check Status")

    def run(self):
        """Run main class code.

        Get data from docker-compose, set instance properties using
        yaml configuration file and start a infinite loop for the dashboard.

        Stops at Error or typing CTRL-C.
        """
        try:
            self.get_compose_data()
            self.get_data_from_version()
            term = Terminal()

            with term.fullscreen():
                with term.hidden_cursor():
                    while True:
                        self.check_time()
                        self.included_services = []
                        ui = self.get_layout(term)
                        ui.display()
                        sleep(self.interval)
        except KeyboardInterrupt:
            print(term.color(0))
            sys.exit(0)
        except BaseException as exc:
            console.error("Internal Error: {}".format(exc))
            if client:
                client.captureException()
            sys.exit(1)

    def _get_branch(self, name, box):
        """Get branch data for service.

        Check if service has build context inside docker-compose
        Try to get git branch information using build context path
        If success and 'target_branch' parameter is set to current box,
        try to get commit diff between current branch and target branch.

        Args:
            name (string): Service name
            box (dict): Box

        Returns
        -------
            String: branch information

        """
        if self.data['services'][name].get('build'):
            data = self.data['services'][name]['build']
            if isinstance(data, str):
                raw_path = data
            else:
                raw_path = self.data['services'][name]['build']['context']

            path = get_path(raw_path, self.path)

            repo = run_command(
                "cd {} && git branch | grep \"*\" 2>/dev/null".format(path),
                get_stdout=True
            )
            if repo:
                dirty = run_command(
                    "cd {} && git status --porcelain 2>/dev/null".format(path),
                    get_stdout=True
                )
                self.repo = {
                    'name': repo.replace("* ", "").replace("\n", ""),
                    'path': path,
                    'dirty': True if dirty else False
                }
                text = self.repo['name']
                target_branch = box.get("target_branch", False)
                if target_branch:
                    lines_behind = 0
                    behind = run_command(
                        "cd {} && git log {}..{} --oneline 2>/dev/null".format(
                            path, text, target_branch),
                        get_stdout=True
                    )
                    if behind:
                        lines_behind = len(behind.split("\n"))
                    if lines_behind:
                        text += " ({}: ".format(target_branch)
                        if lines_behind > 0:
                            text += formatStr.error(
                                "{} {}".format(
                                    ARROW_DOWN, lines_behind - 1),
                                use_prefix=False
                            )
                        text += ")"
            else:
                text = formatStr.warning(
                    "Branch Not Found", use_prefix=False)
        else:
            text = formatStr.info("Using Image", use_prefix=False)
        return text

    def run_fetch(self):
        """Thread routine for git fetch.

        Run 'git fetch' command.
        Must be called inside a Python Thread instance.

        """
        try:
            run_command(
                "cd {} && git fetch --all -q 2>/dev/null".format(self.repo['path']),
                get_stdout=True
            )
        except KeyboardInterrupt:
            raise KeyboardInterrupt

    def check_time(self):
        """Check time for git refresh data."""
        self.can_check_git = False
        if not self.last_git_check:
            self.last_git_check = datetime.datetime.now()
            self.can_check_git = True
        else:
            now = datetime.datetime.now()
            time_elapsed = now - self.last_git_check
            if time_elapsed.seconds >= self.fetch:
                self.last_git_check = now
                self.can_check_git = True

    def _get_git_status(self, key, box):
        """Retrieve Git Status from cache or command.

        Checks every 'fetch_each' seconds.
        Starts a thread for the 'git fetch --all' command

        Args:
            key (string): Service name
            box (dict): Current box
        """
        if not self.can_check_git:
            return self.cache_git.get(key, "")

        if not self.data['services'][key].get('build', False):
            return ""
        self.repo = None
        text = "{} ".format(self._get_branch(key, box))
        theme = "dark"
        if self.repo:
            theme = "success"
            if self.repo['dirty']:
                theme = "warning"
            t = threading.Thread(target=self.run_fetch)
            t.start()
            ret = run_command(
                "cd {} && git status -bs --porcelain".format(
                    self.repo['path']),
                get_stdout=True
            )
            if "behind" in ret:
                s = re.search(r"behind (\d+)", ret)
                text += formatStr.error(
                    "{} {} ".format(DOWN, s.group(1)),
                    use_prefix=False
                )
                theme = "warning"
        status = formatStr.success(text, theme=theme, use_prefix=False)
        self.cache_git[key] = status
        return status

    def _check_server(self, name):
        """Get service docker status.

        Run 'docker inspect <name>' and retrieve the response dict

        Args:
            name (string): service name

        Returns
        -------
            str: formatted string with status

        """
        found = [
            k
            for k in self.data['services']
            if k == name
        ]
        if not found:
            text = ""
        else:
            ret = run_command(
                'docker inspect {} 2>/dev/null'.format(name),
                get_stdout=True
            )
            if ret:
                try:
                    ret = json.loads(ret)[0]
                    if ret['State'].get('Health', False):
                        if ret['State']['Status'].lower() == 'running':
                            stats = ret['State']['Health']['Status'].title()
                        else:
                            stats = ret['State']['Status'].title()
                        if ret['State']['Running']:
                            if ret['State']['Health']['Status'] == 'healthy':
                                theme = "success"
                            else:
                                if ret['State']['Health']['FailingStreak'] > 3:
                                    theme = "error"
                                else:
                                    theme = "warning"
                        elif ret['State']['Paused']:
                            theme = "warning"
                        elif not ret['State']['Running']:
                            theme = "dark"
                        else:
                            theme = "error"
                    else:
                        stats = ret['State']['Status'].title()
                        if ret['State']['Running']:
                            theme = "success"
                        elif ret['State']['Paused']:
                            theme = "warning"
                        elif not ret['State']['Running']:
                            theme = "dark"
                        else:
                            theme = "error"
                    text = formatStr.info(stats, theme=theme, use_prefix=False)
                except KeyboardInterrupt:
                    raise KeyboardInterrupt
                except BaseException as exc:
                    text = formatStr.error("Error", use_prefix=False)
            else:
                text = formatStr.warning("Not Found", use_prefix=False)
        return text



"""Dashing GUI module.

Based on https://github.com/FedericoCeratto/dashing
"""
from datetime import datetime
from collections import deque, namedtuple
from buzio import formatStr
try:
    unichr
except NameError:
    unichr = chr

# "graphic" elements

border_bl = u'└'
border_br = u'┘'
border_tl = u'┌'
border_tr = u'┐'
border_h = u'─'
border_v = u'│'
hbar_elements = (u"▏", u"▎", u"▍", u"▌", u"▋", u"▊", u"▉")
vbar_elements = (u"▁", u"▂", u"▃", u"▄", u"▅", u"▆", u"▇", u"█")
braille_left = (0x01, 0x02, 0x04, 0x40, 0)
braille_right = (0x08, 0x10, 0x20, 0x80, 0)
braille_r_left = (0x04, 0x02, 0x01)
braille_r_right = (0x20, 0x10, 0x08)

TBox = namedtuple('TBox', 't x y w h')


class Tile(object):
    """Tile Object."""

    def __init__(
        self,
        title=None,
        border_color=None,
        color=0,
        terminal=None,
        main=False
    ):
        """Init class."""
        self.title = title
        self.color = color
        self.border_color = border_color
        self._terminal = terminal
        self._main = main

    def _display(self, tbox, parent):
        raise NotImplementedError

    def _draw_borders(self, tbox):
        # top border
        print(tbox.t.color(self.border_color) + tbox.t.move(tbox.x, tbox.y) +
              border_tl + border_h * (tbox.w - 2) + border_tr)
        # left and right
        for dx in range(1, tbox.h - 1):
            print(tbox.t.move(tbox.x + dx, tbox.y) + border_v)
            print(tbox.t.move(tbox.x + dx, tbox.y + tbox.w - 1) + border_v)
        # bottom
        print(tbox.t.move(tbox.x + tbox.h - 1, tbox.y) + border_bl +
              border_h * (tbox.w - 2) + border_br)

    def _draw_borders_and_title(self, tbox):
        if self.border_color is not None:
            self._draw_borders(tbox)
        if self.title:
            fill_all_width = (self.border_color is None)
            self._draw_title(tbox, fill_all_width)

        if self.border_color is not None:
            return TBox(tbox.t, tbox.x + 1, tbox.y + 1, tbox.w - 2, tbox.h - 2)

        elif self.title is not None:
            return TBox(tbox.t, tbox.x + 1, tbox.y, tbox.w - 1, tbox.h - 1)

        return TBox(tbox.t, tbox.x, tbox.y, tbox.w, tbox.h)

    def _fill_area(self, tbox, char=""):
        usable_height = tbox.h - 1
        usable_width = tbox.w - 1
        line = "{:{c}^{num}}".format('', c=char, num=usable_width)
        for count in range(1, usable_height):
            print(tbox.t.move(tbox.x + count, tbox.y + 1) + line)

    def display(self):
        """Render current tile and its items.

        Recurse into nested splits if any.
        """
        t = self._terminal
        tbox = TBox(t, 0, 0, t.width, t.height - 1)
        self._fill_area(tbox)

        if self._main:
            self._terminal.clear()
        tbox = TBox(t, 0, 0, t.width, t.height - 1)
        self._display(tbox, None)

    def _draw_title(self, tbox, fill_all_width):
        if not self.title:
            return
        margin = int((tbox.w - len(self.title)) / 20)
        col = '' if self.border_color is None else \
            tbox.t.color(self.border_color)
        if fill_all_width:
            title = ' ' * margin + self.title + \
                ' ' * (tbox.w - margin - len(self.title))
            if ":" in title:
                print(tbox.t.move(tbox.x, tbox.y) + col + "".join(title.split(":")[0]))
                print(tbox.t.move(tbox.x + tbox.h - 1, tbox.y + 1) + col + title.split(":")[1])
            else:
                print(tbox.t.move(tbox.x, tbox.y) + col + title)
        else:
            title = ' ' * margin + self.title + ' ' * margin
            if ":" in title:
                print(tbox.t.move(tbox.x, tbox.y + margin) + col + "".join(title.split(":")[0]))
                print(tbox.t.move(tbox.x + tbox.h - 1, tbox.y + 1) + col + title.split(":")[1])
            else:
                print(tbox.t.move(tbox.x, tbox.y + margin) + col + title)


class Split(Tile):
    """Split Base Class."""

    def __init__(self, *items, **kw):
        """Init class."""
        super(Split, self).__init__(**kw)
        self.items = items

    def _display(self, tbox, parent):
        tbox = self._draw_borders_and_title(tbox)

        if not self.items:
            # empty split
            self._fill_area(tbox, ' ')
            return

        if isinstance(self, VSplit):
            item_height = tbox.h // len(self.items)
            item_width = tbox.w
        else:
            item_height = tbox.h
            item_width = tbox.w // len(self.items)

        x = tbox.x
        y = tbox.y
        for i in self.items:
            i._display(TBox(tbox.t, x, y, item_width, item_height),
                       self)
            if isinstance(self, VSplit):
                x += item_height
            else:
                y += item_width

        # Fill leftover area
        if isinstance(self, VSplit):
            leftover_x = tbox.h - x + 1
            if leftover_x > 0:
                self._fill_area(TBox(tbox.t, x, y, tbox.w, leftover_x), ' ')
        else:
            leftover_y = tbox.w - y + 1
            if leftover_y > 0:
                self._fill_area(TBox(tbox.t, x, y, leftover_y, tbox.h), ' ')


class VSplit(Split):
    """Split Subclass for verical split."""

    pass


class HSplit(Split):
    """Split Subclass for horizontal split."""

    pass


class Text(Tile):
    """Text UI widget."""

    def __init__(self, text, color=0, *args, **kw):
        """Init Class."""
        super(Text, self).__init__(**kw)
        self.text = text
        self.color = color

    def _display(self, tbox, parent):
        tbox = self._draw_borders_and_title(tbox)
        for dx, line in enumerate(self.text.splitlines()):
            print(tbox.t.color(self.color) + tbox.t.move(tbox.x + dx, tbox.y) +
                  line + ' ' * (tbox.w - len(line)))
        dx += 1
        while dx < tbox.h:
            print(tbox.t.move(tbox.x + dx, tbox.y) + ' ' * tbox.w)
            dx += 1


class Log(Tile):
    """Logger-like UI widget."""

    def __init__(self, date_format, *args, **kw):
        """Init class."""
        self.date_format = date_format
        self.logs = deque(maxlen=50)
        super(Log, self).__init__(**kw)

    def _display(self, tbox, parent):
        tbox = self._draw_borders_and_title(tbox)
        n_logs = len(self.logs)
        log_range = min(n_logs, tbox.h)
        start = n_logs - log_range
        print(tbox.t.color(self.color))
        for i in range(0, log_range):
            line = self.logs[start + i]
            print(tbox.t.move(tbox.x + i, tbox.y) + line +
                  ' ' * (tbox.w - len(line)))

        if i < tbox.h:
            for i2 in range(i + 1, tbox.h):
                print(tbox.t.move(tbox.x + i2, tbox.y) + ' ' * tbox.w)

    def append(self, msg):
        """Append new messages to Log."""
        self.logs.append(msg)

    def _get_header(self):
        now = datetime.now()
        date_string = now.strftime(self.date_format)
        return date_string

    def info(self, msg):
        """Append info messages to log."""
        self.logs.append(
            "[{}]".format(self._get_header())
        )
        msg = "[INFO]: {}".format(msg)
        self.logs.append(msg)

    def warn(self, msg):
        """Append warning messages to log."""
        self.logs.append(
            "[{}]".format(self._get_header())
        )
        msg = formatStr.warning(
            "[WARNING]: {}".format(msg),
            use_prefix=False
        )
        self.logs.append(msg)


class HGauge(Tile):
    """Horizontal Gauge UI widget."""

    def __init__(self, label=None, val=100, color=2, **kw):
        """Init class."""
        kw['color'] = color
        super(HGauge, self).__init__(**kw)
        self.value = val
        self.label = label

    def _display(self, tbox, parent):
        tbox = self._draw_borders_and_title(tbox)
        if self.label:
            wi = (tbox.w - len(self.label) - 3) * self.value / 100
            v_center = int((tbox.h) * 0.5)
        else:
            wi = tbox.w * self.value / 100.0
        index = int((wi - int(wi)) * 7)
        bar = hbar_elements[-1] * int(wi) + hbar_elements[index]
        print(tbox.t.color(self.color) + tbox.t.move(tbox.x, tbox.y + 1))
        if self.label:
            pad = tbox.w - 1 - len(self.label) - len(bar)
        else:
            pad = tbox.w - len(bar)
        bar += hbar_elements[0] * pad
        # draw bar
        for dx in range(0, tbox.h):
            m = tbox.t.move(tbox.x + dx, tbox.y)
            if self.label:
                if dx == v_center:
                    # draw label
                    print(m + self.label + ' ' + bar)
                else:
                    print(m + ' ' * len(self.label) + ' ' + bar)
            else:
                print(m + bar)


class VGauge(Tile):
    """Vertical Gauge UI widget."""

    def __init__(self, val=100, color=2, **kw):
        """Init class."""
        kw['color'] = color
        super(VGauge, self).__init__(**kw)
        self.value = val

    def _display(self, tbox, parent):
        tbox = self._draw_borders_and_title(tbox)
        nh = tbox.h * (self.value / 100.5)
        print(tbox.t.move(tbox.x, tbox.y) + tbox.t.color(self.color))
        for dx in range(tbox.h):
            m = tbox.t.move(tbox.x + tbox.h - dx - 1, tbox.y)
            if dx < int(nh):
                bar = vbar_elements[-1] * tbox.w
            elif dx == int(nh):
                index = int((nh - int(nh)) * 8)
                bar = vbar_elements[index] * tbox.w
            else:
                bar = ' ' * tbox.w

            print(m + bar)


class ColorRangeVGauge(Tile):
    """Vertical gauge with color map.

    E.g.: green gauge for values below 50, red otherwise:
    colormap=((50, 2), (100, 1))
    """

    def __init__(self, val=100, colormap=(), **kw):
        """Init class."""
        self.colormap = colormap
        super(ColorRangeVGauge, self).__init__(**kw)
        self.value = val

    def _display(self, tbox, parent):
        tbox = self._draw_borders_and_title(tbox)
        nh = tbox.h * (self.value / 100.5)
        filled_element = vbar_elements[-1]
        for thresh, col in self.colormap:
            if thresh > self.value:
                break
        print(tbox.t.move(tbox.x, tbox.y) + tbox.t.color(col))
        for dx in range(tbox.h):
            m = tbox.t.move(tbox.x + tbox.h - dx - 1, tbox.y)
            if dx < int(nh):
                bar = filled_element * tbox.w
            elif dx == int(nh):
                index = int((nh - int(nh)) * 8)
                bar = vbar_elements[index] * tbox.w
            else:
                bar = ' ' * tbox.w

            print(m + bar)


class VChart(Tile):
    """Vertical chart.

    Values must be between 0 and 100 and can be float.
    """

    def __init__(self, val=100, *args, **kw):
        """Init class."""
        super(VChart, self).__init__(**kw)
        self.value = val
        self.datapoints = deque(maxlen=50)

    def append(self, dp):
        """Append points to chart."""
        self.datapoints.append(dp)

    def _display(self, tbox, parent):
        tbox = self._draw_borders_and_title(tbox)
        filled_element = hbar_elements[-1]
        scale = tbox.w / 100.0
        print(tbox.t.color(self.color))
        for dx in range(tbox.h):
            index = 50 - (tbox.h) + dx
            try:
                dp = self.datapoints[index] * scale
                index = int((dp - int(dp)) * 8)
                bar = filled_element * int(dp) + hbar_elements[index]
                assert len(bar) <= tbox.w, dp
                bar += ' ' * (tbox.w - len(bar))
            except IndexError:
                bar = ' ' * tbox.w
            print(tbox.t.move(tbox.x + dx, tbox.y) + bar)


class HChart(Tile):
    """Horizontal chart, filled."""

    def __init__(self, val=100, *args, **kw):
        """Init class."""
        super(HChart, self).__init__(**kw)
        self.value = val
        self.datapoints = deque(maxlen=500)

    def append(self, dp):
        """Append points to chart."""
        self.datapoints.append(dp)

    def _display(self, tbox, parent):
        tbox = self._draw_borders_and_title(tbox)
        print(tbox.t.color(self.color))
        for dx in range(tbox.h):
            bar = ''
            for dy in range(tbox.w):
                dp_index = - tbox.w + dy
                try:
                    dp = self.datapoints[dp_index]
                    q = (1 - dp / 100) * tbox.h
                    if dx == int(q):
                        index = int((int(q) - q) * 8 - 1)
                        bar += vbar_elements[index]
                    elif dx < int(q):
                        bar += ' '
                    else:
                        bar += vbar_elements[-1]

                except IndexError:
                    bar += ' '

            # assert len(bar) == tbox.w
            print(tbox.t.move(tbox.x + dx, tbox.y) + bar)


class HBrailleChart(Tile):
    """Horizontal Braille Chart UI widget."""

    def __init__(self, val=100, *args, **kw):
        """Init class."""
        super(HBrailleChart, self).__init__(**kw)
        self.value = val
        self.datapoints = deque(maxlen=500)

    def append(self, dp):
        """Append points to chart."""
        self.datapoints.append(dp)

    def _generate_braille(self, l, r):
        v = 0x28 * 256 + (braille_left[l] + braille_right[r])
        return unichr(v)

    def _display(self, tbox, parent):
        tbox = self._draw_borders_and_title(tbox)
        print(tbox.t.color(self.color))
        for dx in range(tbox.h):
            bar = ''
            for dy in range(tbox.w):
                dp_index = (dy - tbox.w) * 2
                try:
                    dp1 = self.datapoints[dp_index]
                    dp2 = self.datapoints[dp_index + 1]
                except IndexError:
                    # no data (yet)
                    bar += ' '
                    continue

                q1 = (1 - dp1 / 100) * tbox.h
                q2 = (1 - dp2 / 100) * tbox.h
                if dx == int(q1):
                    index1 = int((q1 - int(q1)) * 4)
                    if dx == int(q2):  # both datapoints in the same rune
                        index2 = int((q2 - int(q2)) * 4)
                    else:
                        index2 = -1  # no dot
                    bar += self._generate_braille(index1, index2)
                elif dx == int(q2):
                    # the right dot only is in the current rune
                    index2 = int((q2 - int(q2)) * 4)
                    bar += self._generate_braille(-1, index2)
                else:
                    bar += ' '

            print(tbox.t.move(tbox.x + dx, tbox.y) + bar)


class HBrailleFilledChart(Tile):
    """Horizontal Filled Braille Chart UI Widget."""

    def __init__(self, val=100, *args, **kw):
        """Init class."""
        super(HBrailleFilledChart, self).__init__(**kw)
        self.value = val
        self.datapoints = deque(maxlen=500)

    def append(self, dp):
        """Append points to chart."""
        self.datapoints.append(dp)

    def _generate_braille(self, lmax, rmax):
        v = 0x28 * 256
        for l in range(lmax):
            v += braille_r_left[l]
        for r in range(rmax):
            v += braille_r_right[r]
        return unichr(v)

    def _display(self, tbox, parent):
        tbox = self._draw_borders_and_title(tbox)
        print(tbox.t.color(self.color))
        for dx in range(tbox.h):
            bar = ''
            for dy in range(tbox.w):
                dp_index = (dy - tbox.w) * 2
                try:
                    dp1 = self.datapoints[dp_index]
                    dp2 = self.datapoints[dp_index + 1]
                except IndexError:
                    # no data (yet)
                    bar += ' '
                    continue

                q1 = (1 - dp1 / 100.0) * tbox.h
                q2 = (1 - dp2 / 100.0) * tbox.h
                if dx == int(q1):
                    index1 = 3 - int((q1 - int(q1)) * 4)
                elif dx > q1:
                    index1 = 3
                else:
                    index1 = 0
                if dx == int(q2):
                    index2 = 3 - int((q2 - int(q2)) * 4)
                elif dx > q2:
                    index2 = 3
                else:
                    index2 = 0
                bar += self._generate_braille(index1, index2)

            print(tbox.t.move(tbox.x + dx, tbox.y) + bar)



"""Version package.

Checks version number for upgrades in PyPI
"""
import requests
import sys
from buzio import console
from pkg_resources import parse_version
from cabrita import __version__
from cabrita.utils import run_command


def versions():
    """Function: versions.

    Summary: Request all versions registered in PyPI
    Returns: list
    """
    console.info("Checking for updates...")
    url = "https://pypi.python.org/pypi/cabrita/json"
    data = None
    versions = None
    try:
        ret = requests.get(url, timeout=1)
        data = ret.json()
    except BaseException:
        pass
    if data:
        versions = list(data["releases"].keys())
        versions.sort(key=parse_version)
    return versions


def check_version():
    """Function: check_version.

    Summary: Compares actual version vs last known
    version in PyPI for upgrades
    Returns: Bool = true if updated
    """
    last_version = __version__
    version_data = versions()
    if version_data:
        last_version = version_data[-1]
    if parse_version(last_version) > parse_version(__version__) and \
            ("rc" not in last_version and
                "b" not in last_version and "dev" not in last_version):
        console.warning(
            "You're running a outdated version.\n" +
            "Last Version: {}\n".format(last_version)
        )
        ret = console.confirm("Do you want to upgrade?")
        if ret:
            result = run_command("sudo pip3 install -U cabrita")
            if result:
                print("\nOperation complete. Please run again.")
            else:
                print("\nThere is a error during upgrade. Please try again.")
            sys.exit(0)



"""Setup.py."""
import datetime
from codecs import open
from os import path
from setuptools import setup, find_packages
from cabrita import __version__


here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

install_requires = [
    "tabulate",
    "buzio",
    "click",
    "psutil",
    "PyYAML",
    "requests",
    "blessed",
    "raven"
]

if 'dev' in __version__:
    now = datetime.datetime.now()
    release_number = (now - datetime.datetime(2017, 11, 30)
                      ).total_seconds() / 60
    version = "{}{}".format(__version__, int(release_number))
else:
    version = __version__

setup(
    name='cabrita',
    version=version,
    description='Create dashboard from docker-compose.yml',
    long_description=long_description,
    author='Chris Maillefaud',
    include_package_data=True,
    # Choose your license
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: User Interfaces',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    keywords='terminal tui docker compose dashboard',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'cabrita=cabrita.run:run'
        ],
    },
)
